/* sims.c
 *
 * Simulation server main
 */
#include "sim.h"
#include "server.h"

static int read_map_file(char *name);
static int build_server_list(void);
int read_data_file(char *name, char *desc, int (*parsefunc)(int fd));
static void crash_stop(int sig);
static void trap_signals(void);
static void signal_trapper(int sig);

int parse_debug 	= 0;	/* print heaps of stuff while parsing? */
int fd_debug		= 0;	/* print everything send to server/file? */
int is_sim_server	= 1;	/* this is the server */
int dont_dump           = 0;	/* don't do state dumps */

map *mp;			/* global server info */
int serverfd;			/* master socket for server */

int main(int argc, char **argv)
{
char *mapfile=NULL;		/* filename to read map from */
int servport = SIM_PORT;	/* tcp port to listen on */
time_t lastdumptime=time(NULL);	/* last time server state was saved */
time_t lastmetatime = 0;	/* last time metaserver was contacted */
char logfile[256];		/* log file name */
int i;

srand(time(NULL));

/* Parse command line args
 */
for(i=1; i<argc; i++) {
	if (strequ(argv[i],"-parse_debug")) {
		parse_debug = 1;
		}
	else if (strequ(argv[i],"-fd_debug")) {
		fd_debug = 1;
		}
	else if (strequ(argv[i],"-port") && i<argc-1) {
		servport = atoi(argv[++i]);
		}
	else if (strequ(argv[i],"-dont_dump")) {
		dont_dump = 1;
		}
	else {
		mapfile = argv[i];
		}
	}
if (!mapfile) {
	logerror("No map file given");
	exit(1);
	}

/* Open server main server socket to accept new connections on
 */
if ((serverfd = open_tcp_dest(servport)) == -1) {
	logerror("Couldn't open server socket");
	exit(2);
	}

/* Initialise server map
 */
if (!read_map_file(mapfile))
	exit(2);
if (!build_server_list())
	exit(3);

/* Record this server's address info
 */
mp->server   = strdupe(findmyhostname());
mp->ip       = findmyip();
mp->port     = servport;
if (mp->expectedport && mp->port != mp->expectedport) {
	logerror("Server must be run on port %d",mp->expectedport);
	exit(1);
	}
if (mp->expectedhost && !strequ(mp->expectedhost,mp->server)) {
	logerror("Server must be run on host %s",mp->expectedhost);
	exit(1);
	}

/* Create dirs under map dir for player and object data.
 */
init_data_dir("player_data");
init_data_dir("object_data");

/* Setup signal handlers
 */
set_signals(crash_stop);
trap_signals();

/* start logging
 */
sprintf(logfile,"%s/server_log",mp->mapdir);
open_log_file(logfile);

logmessage("Started server on %s port %d",mp->server,mp->port);

/* Main server loop.
 * Wait for a new connection, request from a client or 1 second, whichever
 * comes first.
 */
while(1) {
	fd_set readfds;		/* file descriptors to wait on */
	player *pp;		/* for traversing player list */
	struct timeval tv;	/* 1 second to wait */
	int maxfd = serverfd;

	/* Wait for something to arrive
	 */
	FD_ZERO(&readfds);
	FD_SET(serverfd,&readfds);
	reset_player();
	while((pp = next_online_player())) {
		FD_SET(pp->fd,&readfds);
		if (pp->fd > maxfd)
			maxfd = pp->fd;
		}
	tv.tv_sec = 1;
	tv.tv_usec = 0;
	if (select(maxfd+1,&readfds,NULL,NULL,&tv) == -1) {
		/* interrupted by a signal.. forget it */
		continue;
		}

	/* Find out who sent something
	 */
	reset_player();
	while((pp = next_online_player())) {
		if (FD_ISSET(pp->fd,&readfds))
			process_request(pp);
		}
	if (FD_ISSET(serverfd,&readfds))
		process_connection(serverfd);

	/* Clean up finished child processes
	 */
	reap_processes();

	/* Do stuff that has to be done occasionally, like controlling NPCs
	 * or fading out players offline for >5 days
	 */
	control_others();
	check_old_players();

	/* Dump server state to a file every 5 minutes
	 */
	if (time(NULL) - lastdumptime > 5*60) {
		lastdumptime = time(NULL);
		dump_state();
		}

	/* Contact the metaserver once per hour
	 */
	if (time(NULL) - lastmetatime > 60*60) {
		lastmetatime = time(NULL);
		contact_metaserver();
		}
	}
}


/* read_map_file
 * Read map file specifing other files to read.
 * Read nodes, paths, shapes, etc..
 * Returns 1 if successful, 0 otherwise
 */
static int read_map_file(char *name)
{
int fd, pfd, pid, i;
person *pr;
player *pl;
object *ob;

/* Allocate and init map structure
 */
mp = scalloc(1,sizeof(map));
mp->name     = "no name";
mp->creator  = "noone";
mp->password = "foo";
mp->safenodec = 0;
mp->mapdir   = strdupe(dirnam(name));
mp->mapfile  = strdupe(basenam(name));
mp->playerc  = 0;

/* Read map file
 */
if ((fd = open(name,O_RDONLY)) == -1) {
	logerror("Couldn't open map file");
	return 0;
	}
if (parse_map((pfd = pipethrough(fd,&pid,CPPCMD)),mp) || pipeclose(pfd)) {
	logerror("Error parsing map");
	kill(pid,SIGKILL);
	return 0;
	}

if (!read_data_file(mp->ndfile,"nodes",parse_nodes))
	return 0;

if (!read_data_file(mp->ptfile,"paths",parse_paths))
	return 0;

if (!read_data_file(mp->shfile,"shapes",parse_shapes))
	return 0;

if (mp->prfile && !read_data_file(mp->prfile,"persons",parse_persons))
	return 0;

if (mp->obfile && !read_data_file(mp->obfile,"objects",parse_objects))
	return 0;

if (mp->plfile && !read_data_file(mp->plfile,"players",parse_players))
	return 0;

if (mp->acfile && !read_data_file(mp->acfile,"actions",parse_actions))
	return 0;

/* Check for safenode and default player shape
 */
if (mp->safenodec == 0) {
	logerror("No safe nodes defined");
	return 0;
	}
for(i=0; i<mp->safenodec; i++) {
	if (!lookup_node(mp->safenode[i])) {
		logerror("Safe node %s does not exist",mp->safenode[i]);
		return 0;
		}
	}
if (!lookup_shape("_player")) {
	logerror("Default player shape \"_player\" not defined");
	return 0;
	}

/* Make sure all objects have an associated shape, and are either in someone's
 * inventory or at a node.
 */
reset_object();
while((ob = next_object())) {
	if (!lookup_shape(ob->name)) {
		logerror("Object \"%s\" has no shape",ob->name);
		return 0;
		}
	if (!ob->pos && !ob->inv) {
		logerror("Object \"%s\" is not at a node or in an inventory",
			 ob->name);
		return 0;
		}
	}

/* Make sure all persons have an associated shape
 */
reset_person();
while((pr = next_person()))
	if (!lookup_shape(pr->name)) {
		logerror("Person \"%s\" has no shape",pr->name);
		return 0;
		}

/* Make sure all players have an associated person
 */
reset_player();
while((pl = next_player()))
	if (!lookup_person(pl->name)) {
		logerror("Player \"%s\" has no person",pl->name);
		return 0;
		}

/* Check that person shapes are where they should be
 */
reset_person();
while((pr = next_person())) {
	shape *psh = lookup_shape(pr->name);

	if (psh->pos.x != pr->pos->x ||
	    psh->pos.y != pr->pos->y ||
	    psh->pos.z != pr->pos->z)
		logerror("Warning! Possible position mismatch for person %s",
			 pr->name);
	}

/* Fill in empty fields in map structure
 */
if (!mp->obfile)
	mp->obfile = strdupe("objects");
if (!mp->prfile)
	mp->prfile = strdupe("persons");
if (!mp->plfile)
	mp->plfile = strdupe("players");
if (!mp->acfile)
	mp->acfile = strdupe("actions");

/* Make sure all players are transparent
 */
reset_person();
while((pr = next_person()))
	if (lookup_player(pr->name)) {
		shape *psh = lookup_shape(pr->name);

		shade_shape(psh,sft_glass);
		}

return 1;
}


/* build_server_list
 * Creates a list of all servers this one is connected to by scanning nodes
 * table. Returns 1 if ok, or 0 if something went wrong.
 */
static int build_server_list(void)
{
node *n;

reset_node();
while((n = next_node())) {
	if (n->special) {
		simserver **sp;
		struct hostent *hent;

		if ((hent = gethostbyname(n->serv.addr)) == NULL) {
			logerror("Connected server %s does not exist!",
				 n->serv.addr);
			return 0;
			}
		n->serv.ip = *(struct in_addr *)hent->h_addr;

		for(sp=&mp->shead; *sp; sp = &((*sp)->next))
			if (strequ((*sp)->addr,n->serv.addr) &&
			    (*sp)->port == n->serv.port)
				break;	/* already in list */

		if (!*sp) {
			/* Not already in list
			 */
			(*sp) = smalloc(sizeof(simserver));
			(*sp)->addr = strdupe(n->serv.addr);
			(*sp)->ip   = n->serv.ip;
			(*sp)->port = n->serv.port;
			(*sp)->next = NULL;
			}
		}
	}

return 1;
}


/* read_data_file
 * General function for checking, opening and parsing a sim data file.
 * Returns 1 if successful, or 0 on error
 */
int read_data_file(char *name, char *desc, int (*parsefunc)(int fd))
{
char fullpath[256], cppcmd[256];
int fd, pfd, pid;

if (!name) {
	logerror("%s file not given",desc);
	return 0;
	}

/* Nodes, shapes, etc.. files are always relative to the map dir
 */
sprintf(fullpath,"%s/%s",mp->mapdir,name);
logmessage("Reading %s file %s",desc,fullpath);
if ((fd = open(fullpath,O_RDONLY)) == -1) {
	logerror("couldn't open file %s",fullpath);
	return 0;
	}

sprintf(cppcmd,"%s -I%s -I%s",CPPCMD,mp->mapdir,dirnam(mp->mapdir));
if (parsefunc((pfd = pipethrough(fd,&pid,cppcmd))) || pipeclose(pfd)) {
	logerror("error parsing file %s",fullpath);
	kill(pid,SIGKILL);
	close(fd);
	return 0;
	}
close(fd);
return 1;
}


/* crash_stop
 * Called when the server is killed
 */
static void crash_stop(int sig)
{
int dc;

set_signals(SIG_IGN);	/* don't let anything iterfere with dumping state */
logmessage("Server killed with signal %d. Shutting down.",sig);
dc = disconnect_players();
if (dc) logmessage("Disconnected %d players",dc);
close(serverfd);
dump_state();
if (!dont_dump) logmessage("Done dumping server state to %s",mp->mapdir);
exit(10);
}


/* trap_signals
 * Setup trapping for annoying signals
 */
static void trap_signals(void)
{
signal(SIGPIPE,signal_trapper);
}


/* signal_trapper
 * Handler for ingnoring signals
 */
static void signal_trapper(int sig)
{
trap_signals();
logmessage("Server caught signal %d.",sig);
}

